import pandas as pd
import numpy as np
from functools import reduce
import warnings


def fair_from_dict(fair_dict, label=["Curr_perf", "Curr_Perc_label"]):
    """Generates a Multiindex Dataframe starting from current fairness metrics dictionary.

    Args:
        fair_dict (dict): dictionary containing current fairness metrics performance
        label (list, optional): label names of column at level 2 in the Multiindex DataFrame. Deafaults to ["Curr_perf","Curr_Perc_label"].

    Returns:
        pd.DataFrame: Multiindex Dataframe
    """
    reformed_total = dict()
    for a in fair_dict.keys():
        reformed_dict = dict()
        for outerKey, innerDict in fair_dict[a].items():
            for innerKey, values in innerDict.items():
                for i in range(len(label)):
                    reformed_dict[(outerKey, innerKey, label[i])] = values[i]
        reformed_total[a] = reformed_dict

    return pd.DataFrame.from_dict(reformed_total, orient="index").reset_index().rename(columns={"index": "Metric"})


def add_absolute_alert_columns(report_df, absolute=False):
    """Add empty columns of Absolute warning for each label (or group of labels) for each feature (or group of features) to the Multiindex Report of FairnessDrift Class.

    Args:
        report_df (pd.DataFrame): Multiindex report of the FairnessDrift Class.
        absolute (bool, optional): it indicates if the columns on absolute warning have been already added. Defaults to False.

    Returns:
        pd.DataFrame: Multiindex DataFrame with Absolute warning columns added
    """
    if not absolute:
        old_dict = report_df.to_dict()
        new_dict = {("Metric", "", ""): old_dict[("Metric", "", "")]}
        for x in [y for y in old_dict.keys() if y != ("Metric", "", "")]:
            new_dict[x] = old_dict[x]
            if x[2] == "Curr_Perc_label":
                new_dict[(x[0], x[1], "Absolute_warning")] = {k: np.nan for k in old_dict[x].keys()}
        return pd.DataFrame(new_dict)
    else:
        return report_df


def add_relative_alert_columns(report_df):
    """Add empty columns of Relative warning and drift for each label (or group of labels) for each feature (or group of features) to the Multiindex Report of FairnessDrift Class.

    Args:
        report_df (pd.DataFrame): Multiindex report of the FairnessDrift Class.

    Returns:
        pd.DataFrame: Multiindex DataFrame with Relative warning and drift columns added
    """
    old_dict = report_df.to_dict()
    new_dict = {("Metric", "", ""): old_dict[("Metric", "", "")]}
    for x in [y for y in old_dict.keys() if y != ("Metric", "", "")]:
        new_dict[x] = old_dict[x]
        if x[2] == "Stor_Perc_label":
            new_dict[(x[0], x[1], "Drift(%)")] = {k: np.nan for k in old_dict[x].keys()}
            new_dict[(x[0], x[1], "Relative_warning")] = {k: np.nan for k in old_dict[x].keys()}

    return pd.DataFrame(new_dict)


def check_fairness_groups(df_1, df_2):
    """Check if two Multiindex Dataframes have different columns until level 1 and retrieves the list of unmatched fairness groups unmatched.

    Args:
        df_1 (pd.DataFrame): Multiindex DataFrame 1
        df_2 (pd.DataFrame): Multiindex DataFrame 2

    Returns:
        list: list of unmatched fairness groups unmatched.
    """
    col_2_train = reduce(lambda l, y: l.append(y) or l if y not in l else l, [(x[0], x[1]) for x in df_1.columns], [])
    col_2_test = reduce(lambda l, y: l.append(y) or l if y not in l else l, [(x[0], x[1]) for x in df_2.columns], [])
    list_no_join = list(set(col_2_train) ^ set(col_2_test))
    if len(list_no_join) != 0:
        warnings.warn(f"unmatched fairness groups {list_no_join}")

    return list_no_join


def bg_red_yellow(df):
    """Color report of Class Fairness Drift based on Alert.

    Args:
        df (pd.DataFrame): report in input

    Returns:
        pd.DataFrame: color-mapping report
    """
    ret = pd.DataFrame("", index=df.index, columns=df.columns)
    for x in np.unique([x for x in df.columns.levels[0] if x != "Metric"]):
        for y in np.unique(df[x].columns.get_level_values(0)):
            if "Absolute_warning" in df[x][y].columns:
                ret.loc[
                    df[x][y].Absolute_warning == "Red Alert", [(x, y, "Curr_perf"), (x, y, "Absolute_warning")]
                ] = "background-color: red"
                ret.loc[
                    df[x][y].Absolute_warning == "Yellow Alert", [(x, y, "Curr_perf"), (x, y, "Absolute_warning")]
                ] = "background-color: yellow"

            ret.loc[
                df[x][y].Relative_warning == "Red Alert", [(x, y, "Drift(%)"), (x, y, "Relative_warning")]
            ] = "background-color: red"
            ret.loc[
                df[x][y].Relative_warning == "Yellow Alert", [(x, y, "Drift(%)"), (x, y, "Relative_warning")]
            ] = "background-color: yellow"
    return ret



import numpy as np
from functools import reduce
import warnings
from typing import Dict

from model_monitoring.utils import check_metrics_sets
from model_monitoring.config import read_config

from model_monitoring.fairness_drift.fairness_drift import (
    fair_from_dict,
    add_absolute_alert_columns,
    add_relative_alert_columns,
    check_fairness_groups,
)

standard_threshold = read_config(config_dir="config", name_params="fairness_drift_threshold.yml")


class FairnessDrift:
    """Fairness Drift Class."""

    def __init__(self, fair_metrics_curr, config_threshold=None):
        """Fairness Drift Class.

        Args:
            fair_metrics_curr (dict): dictionary containing current fairness metrics perfomances
            config_threshold (dict, optional): dictionary containing threshold settings. Defaults to None.
        """
        if not isinstance(fair_metrics_curr, Dict):
            raise ValueError(
                "Fairness metrics in input has not a valid format. It should be a dictionary containing functions as keys and values as values."
            )
        if config_threshold is None:
            config_threshold = standard_threshold

        check_metrics = [i for i in fair_metrics_curr.keys() if i not in config_threshold.keys()]
        if len(check_metrics) > 0:
            warnings.warn(f"{check_metrics} do not have threshold settings in config_threshold")

        list_com_metrics = list(set(fair_metrics_curr.keys()).intersection(set(config_threshold.keys())))

        # initialize report
        report_df = fair_from_dict(
            {x: fair_metrics_curr[x] for x in list_com_metrics}, label=["Curr_perf", "Curr_Perc_label"]
        )

        # for output report columns ordering according to current percentage label
        ordered_columns = (
            report_df.loc[:, (slice(None), slice(None), "Curr_Perc_label")]
            .max()
            .reset_index()
            .sort_values(["level_0", 0], ascending=[True, False])
            .values
        )
        self.report = report_df.reindex([""] + [x[1] for x in ordered_columns], axis=1, level=1)

        self.fair_metrics_curr = fair_metrics_curr
        self.config_threshold = config_threshold
        self.perf_metrics_stor = None
        self.absolute = False
        self.relative = False

    def get_absolute(self):
        """Load on the report the absolute alert on current fairness metrics perfomances."""
        # Regenerate Multiindex report with Absolute_warning columns
        self.report = add_absolute_alert_columns(self.report, absolute=self.absolute)

        # Generation Alert
        if not self.absolute:
            for a in self.report.Metric.values:
                absolute_red = self.config_threshold[a]["absolute"]["red"]
                absolute_yellow = self.config_threshold[a]["absolute"]["yellow"]

                if self.config_threshold[a]["logic"] == "decrease":
                    if absolute_red != "None":
                        for x in np.unique([x for x in self.report.columns.levels[0] if x != "Metric"]):
                            for y in np.unique(self.report[x].columns.get_level_values(0)):
                                curr_perf = self.report.loc[self.report.Metric == a, x][y]["Curr_perf"].values[0]
                                if curr_perf < absolute_red:
                                    self.report.loc[self.report.Metric == a, (x, y, "Absolute_warning")] = "Red Alert"
                                else:
                                    if absolute_yellow != "None":
                                        if (curr_perf > absolute_red) and (curr_perf < absolute_yellow):
                                            self.report.loc[
                                                self.report.Metric == a, (x, y, "Absolute_warning")
                                            ] = "Yellow Alert"
                    else:
                        if absolute_yellow != "None":
                            for x in np.unique([x for x in self.report.columns.levels[0] if x != "Metric"]):
                                for y in np.unique(self.report[x].columns.get_level_values(0)):
                                    curr_perf = self.report.loc[self.report.Metric == a, x][y]["Curr_perf"].values[0]
                                    if curr_perf < absolute_yellow:
                                        self.report.loc[
                                            self.report.Metric == a, (x, y, "Absolute_warning")
                                        ] = "Yellow Alert"

                elif self.config_threshold[a]["logic"] == "increase":
                    if absolute_red != "None":
                        for x in np.unique([x for x in self.report.columns.levels[0] if x != "Metric"]):
                            for y in np.unique(self.report[x].columns.get_level_values(0)):
                                curr_perf = self.report.loc[self.report.Metric == a, x][y]["Curr_perf"].values[0]
                                if curr_perf > absolute_red:
                                    self.report.loc[self.report.Metric == a, (x, y, "Absolute_warning")] = "Red Alert"
                                else:
                                    if absolute_yellow != "None":
                                        if (curr_perf < absolute_red) and (curr_perf > absolute_yellow):
                                            self.report.loc[
                                                self.report.Metric == a, (x, y, "Absolute_warning")
                                            ] = "Yellow Alert"
                    else:
                        if absolute_yellow != "None":
                            for x in np.unique([x for x in self.report.columns.levels[0] if x != "Metric"]):
                                for y in np.unique(self.report[x].columns.get_level_values(0)):
                                    curr_perf = self.report.loc[self.report.Metric == a, x][y]["Curr_perf"].values[0]
                                    if curr_perf > absolute_yellow:
                                        self.report.loc[
                                            self.report.Metric == a, (x, y, "Absolute_warning")
                                        ] = "Yellow Alert"

                elif self.config_threshold[a]["logic"] == "axial":
                    if absolute_red != "None":
                        for x in np.unique([x for x in self.report.columns.levels[0] if x != "Metric"]):
                            for y in np.unique(self.report[x].columns.get_level_values(0)):
                                curr_perf = self.report.loc[self.report.Metric == a, x][y]["Curr_perf"].values[0]
                                if (curr_perf > max(absolute_red)) or (curr_perf < min(absolute_red)):
                                    self.report.loc[self.report.Metric == a, (x, y, "Absolute_warning")] = "Red Alert"
                                else:
                                    if absolute_yellow != "None":
                                        if ((curr_perf < max(absolute_red)) and (curr_perf > min(absolute_red))) and (
                                            (curr_perf > max(absolute_yellow)) or (curr_perf < min(absolute_yellow))
                                        ):
                                            self.report.loc[
                                                self.report.Metric == a, (x, y, "Absolute_warning")
                                            ] = "Yellow Alert"
                    else:
                        if absolute_yellow != "None":
                            for x in [x for x in self.report.columns.levels[0] if x != "Metric"]:
                                for y in self.report[x].columns.get_level_values(0):
                                    curr_perf = self.report.loc[self.report.Metric == a, x][y]["Curr_perf"].values[0]
                                    if (curr_perf > max(absolute_yellow)) or (curr_perf < min(absolute_yellow)):
                                        self.report.loc[
                                            self.report.Metric == a, (x, y, "Absolute_warning")
                                        ] = "Yellow Alert"

                else:
                    raise ValueError(
                        f"{self.config_threshold[a]['logic']} is not a valid logic for {a} metric. Choose between ['increase','decrease','axial']."
                    )

        self.absolute = True

        # for output report columns ordering according to current percentage label
        ordered_columns = (
            self.report.loc[:, (slice(None), slice(None), "Curr_Perc_label")]
            .max()
            .reset_index()
            .sort_values(["level_0", 0], ascending=[True, False])
            .values
        )
        self.report = self.report.reindex([""] + [x[1] for x in ordered_columns], axis=1, level=1)

        # for output report columns ordering
        if self.relative:
            self.report = self.report.reindex(
                [
                    "",
                    "Curr_perf",
                    "Curr_Perc_label",
                    "Absolute_warning",
                    "Stor_perf",
                    "Stor_Perc_label",
                    "Drift(%)",
                    "Relative_warning",
                ],
                axis=1,
                level=2,
            )

    def get_relative(self, fair_metrics_stor):
        """Load on the report the historichal fairness metrics performances, drift compared to the current fairness performances and relative alert on drift.

        Args:
            fair_metrics_stor (dict): dictionary containing historichal fairness metrics perfomances.
        """
        # re-initialize report
        if self.relative:
            list_drop = reduce(
                lambda l, y: l.append(y) or l if y not in l else l,
                [
                    (x[0], x[1], x[2])
                    for x in self.report.columns
                    if x[2] in ["Stor_perf", "Stor_Perc_label", "Drift(%)", "Relative_warning"]
                ],
                [],
            )
            self.report = self.report.drop(columns=list_drop)

        # Check if the metrics are the same
        check_metrics_sets(metrics_1=fair_metrics_stor, metrics_2=self.fair_metrics_curr)

        list_com_metrics = list(set(fair_metrics_stor.keys()).intersection(set(self.fair_metrics_curr.keys())))
        self.fair_metrics_stor = {x: fair_metrics_stor[x] for x in list_com_metrics}
        stor_fair_df = fair_from_dict(self.fair_metrics_stor, label=["Stor_perf", "Stor_Perc_label"])

        # Check if the fairness group are the same
        list_no_join = check_fairness_groups(self.report, stor_fair_df)

        # Update the report with historical fairness performances and limit to common fairness groups
        self.report = (
            self.report.merge(stor_fair_df, how="outer", on="Metric")
            .sort_index(axis=1, level=1, ascending=False)[
                reduce(lambda l, y: l.append(y) or l if y not in l else l, [x[0] for x in self.report.columns], [])
            ]
            .drop(columns=list_no_join)
        )

        # Regenerate Multiindex report with Drift and columns
        self.report = add_relative_alert_columns(self.report)

        # Generation Drift
        for a in self.report.Metric.values:
            if self.config_threshold[a]["logic"] in ["decrease", "increase"]:
                for x in np.unique([x for x in self.report.columns.levels[0] if x != "Metric"]):
                    for y in np.unique(self.report[x].columns.get_level_values(0)):
                        stor_perf = self.report.loc[self.report.Metric == a, x][y]["Stor_perf"].values[0]
                        curr_perf = self.report.loc[self.report.Metric == a, x][y]["Curr_perf"].values[0]
                        if stor_perf > 0:
                            self.report.loc[self.report.Metric == a, (x, y, "Drift(%)")] = (
                                (curr_perf - stor_perf) / stor_perf * 100
                            )
                        else:
                            self.report.loc[self.report.Metric == a, (x, y, "Drift(%)")] = (
                                (stor_perf - curr_perf) / stor_perf * 100
                            )
            elif self.config_threshold[a]["logic"] == "axial":
                axial_point = self.config_threshold[a]["axial_point"]
                for x in np.unique([x for x in self.report.columns.levels[0] if x != "Metric"]):
                    for y in np.unique(self.report[x].columns.get_level_values(0)):
                        stor_perf = self.report.loc[self.report.Metric == a, x][y]["Stor_perf"].values[0]
                        curr_perf = self.report.loc[self.report.Metric == a, x][y]["Curr_perf"].values[0]
                        self.report.loc[self.report.Metric == a, (x, y, "Drift(%)")] = (
                            (abs(curr_perf - axial_point) - abs(stor_perf - axial_point))
                            / abs(stor_perf - axial_point)
                            * 100
                        )
            else:
                raise ValueError(
                    f"{self.config_threshold[a]['logic']} is not a valid logic for {a} metric. Choose between ['increase','decrease','axial']."
                )

        # Generation Alert
        for a in self.report.Metric.values:
            relative_red = self.config_threshold[a]["relative"]["red"]
            relative_yellow = self.config_threshold[a]["relative"]["yellow"]
            absolute_tol = self.config_threshold[a]["relative"]["absolute_tol"]

            if self.config_threshold[a]["logic"] == "decrease":
                if relative_red != "None":
                    for x in np.unique([x for x in self.report.columns.levels[0] if x != "Metric"]):
                        for y in np.unique(self.report[x].columns.get_level_values(0)):
                            # check absolute tollerance for relative alert
                            stor_perf = self.report.loc[self.report.Metric == a, x][y]["Stor_perf"].values[0]
                            curr_perf = self.report.loc[self.report.Metric == a, x][y]["Curr_perf"].values[0]
                            if abs(curr_perf - stor_perf) >= absolute_tol:
                                drift_perf = self.report.loc[self.report.Metric == a, x][y]["Drift(%)"].values[0]
                                if drift_perf < relative_red * 100:
                                    self.report.loc[self.report.Metric == a, (x, y, "Relative_warning")] = "Red Alert"
                                else:
                                    if relative_yellow != "None":
                                        if (drift_perf > relative_red * 100) and (drift_perf < relative_yellow * 100):
                                            self.report.loc[
                                                self.report.Metric == a, (x, y, "Relative_warning")
                                            ] = "Yellow Alert"
                else:
                    if relative_yellow != "None":
                        for x in np.unique([x for x in self.report.columns.levels[0] if x != "Metric"]):
                            for y in np.unique(self.report[x].columns.get_level_values(0)):
                                # check absolute tollerance for relative alert
                                stor_perf = self.report.loc[self.report.Metric == a, x][y]["Stor_perf"].values[0]
                                curr_perf = self.report.loc[self.report.Metric == a, x][y]["Curr_perf"].values[0]
                                if abs(curr_perf - stor_perf) >= absolute_tol:
                                    drift_perf = self.report.loc[self.report.Metric == a, x][y]["Drift(%)"].values[0]
                                    if drift_perf < relative_yellow * 100:
                                        self.report.loc[
                                            self.report.Metric == a, (x, y, "Relative_warning")
                                        ] = "Yellow Alert"

            elif self.config_threshold[a]["logic"] in ["increase", "axial"]:
                if relative_red != "None":
                    for x in np.unique([x for x in self.report.columns.levels[0] if x != "Metric"]):
                        for y in np.unique(self.report[x].columns.get_level_values(0)):
                            # check absolute tollerance for relative alert
                            stor_perf = self.report.loc[self.report.Metric == a, x][y]["Stor_perf"].values[0]
                            curr_perf = self.report.loc[self.report.Metric == a, x][y]["Curr_perf"].values[0]
                            if abs(curr_perf - stor_perf) >= absolute_tol:
                                drift_perf = self.report.loc[self.report.Metric == a, x][y]["Drift(%)"].values[0]
                                if drift_perf > relative_red * 100:
                                    self.report.loc[self.report.Metric == a, (x, y, "Relative_warning")] = "Red Alert"
                                else:
                                    if relative_yellow != "None":
                                        if (drift_perf < relative_red * 100) and (drift_perf > relative_yellow * 100):
                                            self.report.loc[
                                                self.report.Metric == a, (x, y, "Relative_warning")
                                            ] = "Yellow Alert"
                else:
                    if relative_yellow != "None":
                        for x in np.unique([x for x in self.report.columns.levels[0] if x != "Metric"]):
                            for y in np.unique(self.report[x].columns.get_level_values(0)):
                                # check absolute tollerance for relative alert
                                stor_perf = self.report.loc[self.report.Metric == a, x][y]["Stor_perf"].values[0]
                                curr_perf = self.report.loc[self.report.Metric == a, x][y]["Curr_perf"].values[0]
                                if abs(curr_perf - stor_perf) >= absolute_tol:
                                    drift_perf = self.report.loc[self.report.Metric == a, x][y]["Drift(%)"].values[0]
                                    if drift_perf > relative_yellow * 100:
                                        self.report.loc[
                                            self.report.Metric == a, (x, y, "Relative_warning")
                                        ] = "Yellow Alert"

            else:
                raise ValueError(
                    f"{self.config_threshold[a]['logic']} is not a valid logic for {a} metric. Choose between ['increase','decrease','axial']."
                )

        self.relative = True

        # for output report columns ordering according to current percentage label
        ordered_columns = (
            self.report.loc[:, (slice(None), slice(None), "Curr_Perc_label")]
            .max()
            .reset_index()
            .sort_values(["level_0", 0], ascending=[True, False])
            .values
        )
        self.report = self.report.reindex([""] + [x[1] for x in ordered_columns], axis=1, level=1)

        # for output report columns ordering
        if self.absolute:
            self.report = self.report.reindex(
                [
                    "",
                    "Curr_perf",
                    "Curr_Perc_label",
                    "Absolute_warning",
                    "Stor_perf",
                    "Stor_Perc_label",
                    "Drift(%)",
                    "Relative_warning",
                ],
                axis=1,
                level=2,
            )
        else:
            self.report = self.report.reindex(
                ["", "Curr_perf", "Curr_Perc_label", "Stor_perf", "Stor_Perc_label", "Drift(%)", "Relative_warning"],
                axis=1,
                level=2,
            )

    def get_report(self):
        """Return the report.

        Returns:
            pd.DataFrame: report of the class.
        """
        return self.report
